import sqlite3


def create_index(conn, table, columns, unique=False):
    unique_str = "UNIQUE" if unique else ""
    index_name = f"{'_'.join(columns)}_index"
    columns_str = ", ".join(columns)
    sql = f"CREATE {unique_str} INDEX IF NOT EXISTS {index_name} ON {table} ({columns_str})"
    conn.execute(sql)


def build_wikidata_index(db_path):
    conn = sqlite3.connect(db_path)
    print('Database connected.')
    print('wikidataentityensitelink entity_id')
    create_index(conn, 'wikidataentityensitelink', ['entity_id'])

    print('wikidataentitylabel entity_id language')
    create_index(conn, 'wikidataentitylabel', ['entity_id', 'language'])

    print('wikidataentitylabel language value')
    create_index(conn, 'wikidataentitylabel', ['language', 'value'])

    print('wikidataentitydescriptions entity_id language')
    create_index(conn, 'wikidataentitydescriptions', ['entity_id', 'language'])

    print('wikidataentityaliases entity_id language')
    create_index(conn, 'wikidataentityaliases', ['entity_id', 'language'])

    print('wikidataclaim source_entity property_id target_entity')
    create_index(conn, 'wikidataclaim', ['source_entity', 'property_id', 'target_entity'])

    print('wikidataclaim property_id target_entity')
    create_index(conn, 'wikidataclaim', ['property_id', 'target_entity'])

    print('wikidataclaim target_entity')
    create_index(conn, 'wikidataclaim', ['target_entity'])

    print('Indexes created.')

    # Close the connection
    conn.close()
    print('Database connection closed.')



from functools import lru_cache
from .db import WikidataEntityEnSiteLink


def create_link(site_link_name: str, lang='en') -> str:
    title = site_link_name.replace(' ', '_')
    return f"https://{lang}.wikipedia.org/wiki/{title}"


@lru_cache(maxsize=1024 * 30)
def lookup_wikilink(entity, default_value=None):
    """
    lookup wikilink by entity id
    :param entity: wikidata entity id
    :return:
    """
    item = WikidataEntityEnSiteLink.get_or_none(
        WikidataEntityEnSiteLink.entity_id == entity,
    )
    if item:
        return create_link(item.title)
    else:
        return default_value



import bz2
import gzip
import json
from dataclasses import dataclass
from tqdm import tqdm
from cmfn import listdir

from playhouse.sqlite_ext import *

from wikimine.db import (
    WikidataEntityEnSiteLink,
    WikidataEntityLabel,
    WikidataEntityDescriptions,
    WikidataEntityAliases,
    WikidataClaim, db_proxy
)


def smart_open(file_path, mode='r'):
    """
    :param file_path: Path to the file to be open.
    :param mode: The mode in which the file should be opened. Default is 'r'.
    :return: The file object for reading or writing.
    """
    if file_path.endswith('.gz'):
        return gzip.open(file_path, mode + 't') if 'b' not in mode else gzip.open(file_path, mode)
    elif file_path.endswith('.bz2'):
        return bz2.open(file_path, mode + 't') if 'b' not in mode else bz2.open(file_path, mode)
    else:
        return open(file_path, mode)


def iter_jsonl(fn):
    """
    :param fn: The file path of the JSONL file to be read
    :return: A generator object that yields each valid JSON object from the file
    """
    with smart_open(fn) as fd:
        for line in fd:
            line = line.strip()
            if line:
                yield json.loads(line)


@dataclass
class ProcessedEntity:
    entity_id: str
    site_link: WikidataEntityEnSiteLink
    labels: list[WikidataEntityLabel]
    descriptions: list[WikidataEntityDescriptions]
    aliases: list[WikidataEntityAliases]
    claims: list[WikidataClaim]


def get_target_id(mainsnak):
    target_entity = None
    if 'datavalue' in mainsnak:
        datavalue = mainsnak['datavalue']
        if datavalue.get('type') == 'wikibase-entityid':
            target_entity = datavalue.get('value', {}).get('id')
    return target_entity


def process_entity(
        entity_dict,
        process_labels=True,
        process_desc=True,
        process_aliases=True,
        process_claims=True
) -> ProcessedEntity:
    """
    process one line of wikidata json dump
    :param entity_dict:
    :return:
    """
    entity_id = entity_dict['id']
    site_link = None

    if entity_dict.get("sitelinks", {}).get("enwiki", {}).get("title", None) is not None:
        site_link = WikidataEntityEnSiteLink(
            entity_id=entity_id,
            title=entity_dict["sitelinks"]['enwiki']['title']
        )

    # Extract labels
    labels = [
        WikidataEntityLabel(
            entity_id=entity_id,
            language=lang,
            value=label_info['value']
        ) for lang, label_info in entity_dict.get('labels', {}).items()
    ] if process_labels else []

    # Extract descriptions
    descriptions = [
        WikidataEntityDescriptions(
            entity_id=entity_id,
            language=lang,
            value=desc_info['value']
        ) for lang, desc_info in entity_dict.get('descriptions', {}).items()
    ] if process_desc else []

    # Extract aliases
    aliases = [
        WikidataEntityAliases(
            entity_id=entity_id,
            language=lang,
            value=alias_info['value']
        ) for lang, alias_list in entity_dict.get('aliases', {}).items()
        for alias_info in alias_list
    ] if process_aliases else []

    # Extract claims
    claims = []

    if process_claims:
        for property_id, claim_list in entity_dict.get('claims', {}).items():
            # if property_id not in ['P279', 'P31']:
            #     continue

            for claim in claim_list:

                if claim['rank'] == 'deprecated':
                    continue

                mainsnak = claim.get('mainsnak', {})
                target_entity = get_target_id(mainsnak)

                claims.append(
                    WikidataClaim(
                        source_entity=entity_id,
                        property_id=property_id,
                        body=claim,
                        target_entity=target_entity
                    )
                )

    return ProcessedEntity(
        entity_id=entity_id,
        site_link=site_link,
        labels=labels,
        descriptions=descriptions,
        aliases=aliases,
        claims=claims
    )


def has_english_wikipedia_link(entity):
    sitelinks = entity.get("sitelinks", {})
    return "enwiki" in sitelinks


def has_english_label(entity):
    labels = entity.get("labels", {})
    return "en" in labels


def flatten_batch(batch: list[ProcessedEntity]) -> list[WikidataEntityLabel]:
    labels = []
    descriptions = []
    aliases = []
    claims = []
    site_links = []

    for entity in batch:
        labels.extend(entity.labels)
        descriptions.extend(entity.descriptions)
        aliases.extend(entity.aliases)
        claims.extend(entity.claims)
        if entity.site_link:
            site_links.append(entity.site_link)

    return labels, descriptions, aliases, claims, site_links


def insert_wikidata_batch(batch_results, batch_size):
    labels, descriptions, aliases, claims, site_links = flatten_batch(batch_results)
    WikidataEntityLabel.bulk_create(labels, batch_size)
    WikidataEntityDescriptions.bulk_create(descriptions, batch_size)
    WikidataEntityAliases.bulk_create(aliases, batch_size)
    WikidataClaim.bulk_create(claims, batch_size)
    WikidataEntityEnSiteLink.bulk_create(site_links, batch_size)


def add_wikidata_dump_folder(dump_folder, batch_size=5000, dry_run=False, dry_run_limit=10):
    fs = list_wikidata_parts_file(dump_folder)
    if dry_run:
        counter = 0
        ret = []
        for f in tqdm(fs):
            for item in iter_jsonl(f):
                counter += 1
                if counter >= dry_run_limit:
                    return ret
                ret.append(process_entity(item))
        return ret
    else:
        batch_results = []
        with tqdm(desc="Processing items", unit=" item ", mininterval=0.5) as pbar:
            for f in fs:
                pbar.set_description(f"Processing file: {f}")
                for item in iter_jsonl(f):
                    pbar.update(1)
                    entity_id = item['id']
                    not_p = not entity_id.startswith('P')

                    process_labels = True
                    process_desc = True
                    process_aliases = True
                    process_claims = True

                    if not_p and not has_english_wikipedia_link(item):
                        process_labels = has_english_label(item)
                        process_desc = False
                        process_aliases = False
                        process_claims = False

                    batch_results.append(
                        process_entity(
                            item,
                            process_labels=process_labels,
                            process_desc=process_desc,
                            process_aliases=process_aliases,
                            process_claims=process_claims
                        )
                    )
                    if len(batch_results) == batch_size:
                        with db_proxy.atomic():
                            insert_wikidata_batch(batch_results, batch_size)
                            batch_results.clear()

        if len(batch_results) > 0:
            # TODO: insert everything into database
            with db_proxy.atomic():
                insert_wikidata_batch(batch_results, batch_size)
            batch_results.clear()


def list_wikidata_parts_file(dump_folder):
    fs = listdir(dump_folder)
    fs = [f for f in fs if f.endswith('.jsonl') or f.endswith('.jsonl.bz2')]
    return sorted(fs)


def init_wikidata():
    db_proxy.create_tables(
        [WikidataEntityLabel, WikidataEntityDescriptions, WikidataEntityAliases, WikidataClaim,
         WikidataEntityEnSiteLink])



class TypingRelations:
    instance_of = 'P31'
    subclass_of = 'P279'


class People:
    gender = 'P21'

    place_of_birth = 'P19'
    place_of_death = 'P20'

    date_of_birth = 'P569'
    date_of_death = 'P570'

    native_lang = 'P103'
    lang_spoken = 'P1412'
    lang_written = 'P6886'

    occupation = 'P106'
    ethnic_group = 'P172'

    influenced_by = 'P737'
    field_of_work = 'P101'
    educated_at = 'P69'
    notable_work = 'P800'


class WrittenWorks:
    publication_date = 'P577'
    language = 'P407'
    country_of_origin = 'P495'
    main_sub = 'P921'
    genre = 'P136'
    author = 'P50'


class Location:
    country = 'P17'
    coordinate = 'P625'



# SPDX-FileCopyrightText: 2024-present Hao Wu <haowu@dataset.sh>
#
# SPDX-License-Identifier: MIT
from .db import auto_connect, connect
from .labels import lookup_label
from .wiki_link import lookup_wikilink


def is_empty(value):
    if value is None:
        return True
    if not value:
        return True
    if value == '':
        return True

    return False


def list_static_class_members(clz):
    return {(k, v) for k, v in vars(clz).items() if not callable(v) and not k.startswith("__")}



from .db import WikidataClaim
from .labels import lookup_label


class Aggregator(object):

    @staticmethod
    def first_or_none(vs):
        if len(vs) == 0:
            return None
        return vs[0]

    @staticmethod
    def default_agg(x):
        return x

    @staticmethod
    def min(vs):
        if len(vs) == 0:
            return None
        return min(vs)

    @staticmethod
    def max(vs):
        if len(vs) == 0:
            return None
        return max(vs)


class ClaimExtractor(object):

    @staticmethod
    def get_target_label(t):
        return lookup_label(t.target_entity)

    @staticmethod
    def get_target_label_and_id(t):
        if t.target_entity is None:
            return None
        eid = t.target_entity
        label = lookup_label(eid)
        return {
            'id': eid,
            'label': label
        }

    @staticmethod
    def get_target_value(claim):
        return claim.body

    @staticmethod
    def get_target_value(claim):
        return (
            claim.body
            .get('mainsnak', {})
            .get('datavalue', {})
            .get('value', None)
        )

    @staticmethod
    def get_target_value_year(claim):
        time_value = (
            claim.body
            .get('mainsnak', {})
            .get('datavalue', {})
            .get('value', {})
            .get('time')
        )

        if time_value:
            return int(time_value[:5])
        return None


def gather(source_entity, prop_id, map_fn, agg_fn):
    claims = WikidataClaim.select().where(
        WikidataClaim.source_entity == source_entity,
        WikidataClaim.property_id == prop_id,
    )

    if map_fn is None:
        map_fn = ClaimExtractor.get_target_label

    if agg_fn is None:
        agg_fn = Aggregator.default_agg

    processed_claims = []

    for c in claims:
        v = map_fn(c)
        if v:
            processed_claims.append(v)

    return agg_fn(processed_claims)


def fetch(name, prop_id, map_fn=None, agg_fn=None):
    return (name, prop_id, map_fn, agg_fn)


def gather_all(
        source_entity,
        cfg,
        name_field='name',
        entity_id_field='entity_id',
):
    item = {}
    for (name, prop_id, map_fn, agg_fn) in cfg:
        item[name] = gather(source_entity, prop_id, map_fn, agg_fn)

    item[name_field] = lookup_label(source_entity)
    item[entity_id_field] = source_entity
    return item



import json
import os

from peewee import *
from playhouse.sqlite_ext import SqliteExtDatabase, JSONField

# db = SqliteExtDatabase(':memory:')

db_proxy = DatabaseProxy()


class BaseModel(Model):
    class Meta:
        database = db_proxy


class WikidataEntityEnSiteLink(BaseModel):
    entity_id = CharField()
    title = CharField()


class WikidataEntityLabel(BaseModel):
    """
    wikidata entity label
    """
    entity_id = CharField()
    language = CharField()
    value = CharField()


class WikidataEntityDescriptions(BaseModel):
    """
    wikidata entity descriptions.
    example: (Q9191, 'en', 'René Descartes')
    """
    entity_id = CharField()
    language = CharField()
    value = CharField()


class WikidataEntityAliases(BaseModel):
    """
    wikidata entity aliases.
    example:
    (Q9191, 'en', 'Descartes')
    (Q9191, 'en', 'Cartesius')
    """
    entity_id = CharField()
    language = CharField()
    value = CharField()


class WikidataClaim(BaseModel):
    """
    wikidata claim contains Statements about wikidata items.
    (item, property, value).

    You can read more about this concept here:
    https://www.wikidata.org/wiki/Wikidata:Introduction

    This table is indexed by
        (source_entity, property_id, target_entity).
        (property_id, target_entity).
        (target_entity).
    """
    source_entity = CharField()  # entity id of item.
    property_id = CharField()
    body = JSONField()  # this is the claim body.
    target_entity = CharField(null=True)  # this is only true if mainsnak.datavalue is wikibase-entityid


def connect(path_to_db):
    database = SqliteExtDatabase(path_to_db, pragmas=(
        ('cache_size', -1024 * 64),  # 64MB page-cache.
        ('journal_mode', 'wal'),  # Use WAL-mode (you should always use this!).
        ('foreign_keys', 1)))  # Enforce foreign-key constraints.

    db_proxy.initialize(database)


def auto_connect():
    """
    Search for database path from the following source and connect to it automatically.
    1.  from environment variable [WIKIMINE_WIKIDATA_DB].
    2.  ~/.wikimine.config.json: {"db_path": "/path/to/db"}
    :return:
    """
    db_path = os.getenv('WIKIMINE_WIKIDATA_DB')
    if not db_path:
        cfg_file = os.path.expanduser('~/.wikimine.config.json')
        if os.path.exists(cfg_file):
            with open(cfg_file, 'r') as f:
                cfg = json.load(f)
                db_path = cfg['db_path']

    if db_path:
        connect(db_path)
    else:
        raise EnvironmentError('WIKIMINE_WIKIDATA_DB not set')



from functools import lru_cache
from .db import WikidataEntityLabel


@lru_cache(maxsize=1024 + 30)  # Adjust size as needed
def lookup_label(entity, lang='en'):
    item = WikidataEntityLabel.get_or_none(
        WikidataEntityLabel.entity_id == entity,
        WikidataEntityLabel.language == lang,
    )
    return item.value if item else entity



import bz2
import json
import os
from tqdm import tqdm


def split_wikidata_dump(input_file, output_folder, batch_size=1000, stop_at=-1):
    """
    Split a Wikidata JSON.bz2 dump into smaller JSONL.bz2 files.

    Parameters:
    - input_file: str, path to the input Wikidata JSON.bz2 dump file.
    - output_prefix: str, prefix for the output files.
    - batch_size: int, number of entities in each batch file.
    """

    os.makedirs(output_folder, exist_ok=True)

    def write_batch_to_file(batch, batch_num):
        """Write a batch of entities to a compressed JSONL.bz2 file."""
        output_file = os.path.join(output_folder, f"{batch_num:012d}.jsonl.bz2")
        with bz2.open(output_file, 'wt') as bz2_file:
            for entity in batch:
                bz2_file.write(json.dumps(entity) + '\n')

    batch = []
    batch_num = 0

    count = 0

    with bz2.open(input_file, 'rt') as bz2_file:
        for line in tqdm(bz2_file):
            if line.startswith('[') or line.startswith(']'):
                # Skip the starting and ending brackets of the JSON array
                continue
            count += 1

            if stop_at >= 0 and count > stop_at:
                break

            # Remove the trailing comma if it's not the last line
            if line.endswith(',\n'):
                line = line[:-2] + '\n'

            # Parse the line as JSON
            entity = json.loads(line)

            batch.append(entity)

            # If batch size is reached, write the batch to a file
            if len(batch) >= batch_size:
                write_batch_to_file(batch, batch_num)
                batch = []
                batch_num += 1

    # Write the remaining batch if any
    if batch:
        write_batch_to_file(batch, batch_num)

    print(f"Splitting completed. {batch_num + 1} batch files created.")
